import cocos

cocos.director.director.init(
        #resizable=True,
        #do_not_scale=True,
        width=1200,
        height=700,
        #audio_backend='sdl',
        )

class Objstub:
    def __init__(self,x,y):
        self.x=x
        self.y=y

def get_closest_obj(x,y):
    mouse = Objstub(x,y)
    closest = None
    closest_dist = 20.0
    for cell in RESOURCE_HASHMAP.get((mouse.x,mouse.y)):
        for i in cell:
            dist = obj_dist(i,mouse)
            if (dist < closest_dist):
                closest = i
                closest_dist = dist
    for cell in ANT_HASHMAP.get((mouse.x,mouse.y)):
        for i in cell:
            #if not i.color == RED:
            #    continue
            dist = obj_dist(i,mouse)
            if (dist < closest_dist):
                closest = i
                closest_dist = dist
    return closest

class Map(cocos.layer.Layer):
    is_event_handler = True
    def __init__(self):
        import pyglet
        super(Map, self).__init__()
        sprite = cocos.sprite.Sprite('data/world.png', scale=1)
        #sprite = cocos.sprite.Sprite('data/world2.png', scale=10)
        # max extent?

        sprite.position = 0,0
        self.add(sprite,z=-1)
        self.sprite = sprite

        self.schedule_interval(self.on_tick,0.01)
        self.keys = pyglet.window.key.KeyStateHandler()
        cocos.director.director.window.push_handlers(self.keys)
        self.add(ABATCH)
        self.add(RBATCH)

        self.mx = 0
        self.my = 0

        self.mouse_label = cocos.text.Label(
                '%s %s'%(self.mx,self.my),
                font_name = 'Comic Sans MS',
                font_size = 32,
                color = (0,0,0,255),
                anchor_x = 'center',anchor_y='center'
                )
        self.mouse_label.position = self.x, self.y
        self.add(self.mouse_label,z=1)

        self.fx = self.x
        self.fy = self.y

        for ant in ANTS:
            ABATCH.add(ant.sprite,z=0)

        rect = self.sprite.get_rect()
        self.map_width,self.map_height = rect.width,rect.height
        self.min_x = -self.map_width//2+5
        self.min_y = -self.map_height//2+5

        for k in range(500):
            self.make_resource()

    def on_mouse_press(self, x, y, button, modifiers):
        #print 'click'
        x,y = cocos.director.director.get_virtual_coordinates(x, y)
        x,y = int(x),int(y)
        x = x-self.x
        y = y-self.y
        obj =  get_closest_obj(x, y)
        if isinstance(obj,Ant) and (obj.color == RED) and not isinstance(obj, Queen):
            obj.mutate()
        else:
            red_queen.target = Objstub(x,y)

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        self.dx = scroll_x
        self.dy = scroll_y

        self.fx += self.dx*10.0
        self.fy -= self.dy*10.0

    def on_mouse_motion(self, x, y, dx, dy):
        x,y = cocos.director.director.get_virtual_coordinates(x, y)
        x,y = int(x),int(y)
        self.mx = x-self.x
        self.my = y-self.y

    def get_terrain(self, x, y):
        tx,ty = x-self.min_x,y-self.min_y
        if (tx<0) or (ty<0): 
            return -1
        try:
            return terrain_bins[tx//10][-ty//10]
        except IndexError:
            return -1
        #self.pil_x = x-self.x + self.width//2
        #self.pil_y = y-self.y + self.length//2

    def on_tick(self,dt):
        dt = min((dt,0.1))

        #rect = self.sprite.get_rect()
        #self.map_width,self.map_height = rect.width,rect.height

        import pyglet
        keys = self.keys
        k = pyglet.window.key

        self.dx = self.dy = 0.0

        if keys[k.RIGHT] or keys[k.D]:
            self.dx -= 1.0
        if keys[k.LEFT] or keys[k.A]:
            self.dx += 1.0
        if keys[k.UP] or keys[k.W]:
            self.dy -= 1.0
        if keys[k.DOWN] or keys[k.S]:
            self.dy += 1.0

        self.fx += self.dx*dt*3000.0
        self.fy += self.dy*dt*3000.0

        if keys[k.SPACE]:
            self.fx, self.fy = 0,0

        self.x = int(self.fx)
        self.y = int(self.fy)

        self.mouse_label.position = self.mx,self.my+20
        closest = get_closest_obj(self.mx,self.my)
        if closest is None:
            label = bin_name[self.get_terrain(self.mx,self.my)]
        else:
            label = str(closest)
        self.mouse_label.element.text = label#+' %s %s'%(self.mx,self.my)

        for ant in list(ANTS):
            ant.do(dt)
            ant.draw()

        # randomly top up some of the resources
        import random
        n = int(dt*5)
        if random.random() < (dt*5.0-n):
            n+=1
        for k in range(n):
            self.make_resource()

    def make_resource(self):
       # print 'make'
        import random
        x = random.uniform(self.min_x,self.min_x+self.map_width)
        y = random.uniform(self.min_y,self.min_y+self.map_height)
        t = self.get_terrain(int(x),int(y))
        if t>1:
            r = random.random()*2.0
            resource = Resource(x,y,100.0*r*t)
            RBATCH.add(resource.sprite)
            #print '+',100.0*r*t
        #if random.random() < 0.01:
        #    print sum([sum(row) for row in resources])

# 4e3 (4e7 / 1e4) bignodes
# can update in ~ 0.5 seconds.



# bignodes have centroid, neighbours, resource_scent, war scent, neighbour_distance, 
# scent - direction, magnitude.

# 8 directions
# little_nodes - next target-> most likely next step.


# I can hit ~ 10,000 nodes / second.
# I have ~ 350,000
# ~ 35


class Resource:
    def __init__(self,x,y,resources, body = False):
        if body:
            self.sprite = cocos.sprite.Sprite('data/antbody.png')
            self.sprite.scale = 0.5
        else:
            self.sprite = cocos.sprite.Sprite('data/resource.png')

        self.x=x
        self.y=y
        self.resources=resources
        self.sprite.x=x
        self.sprite.y=y
        RESOURCE_HASHMAP.insert((x,y),self)
        self.removed = False

    def __str__(self):
        #return 'Ant: color = %s, life = %.0f, state = %s'%(self.color,self.life,self.state)
        return 'Food: %.0f'%(self.resources)

    def remove(self):
        if not self.removed:
            self.removed = True
            RBATCH.remove(self.sprite)
            RESOURCE_HASHMAP.remove((self.x,self.y),self)


        #self.sprite = 

EXPLORE = 'Exploring'
GET_RESOURCE = 'Getting food'
GO_HOME = 'Going home'
FIGHTING = 'Fighting'
FLEEING = 'Fleeing'

RED = 0
BLUE = 1
GREEN = 2


def obj_dist(obj1,obj2):
    from math import hypot
    return hypot(obj1.x-obj2.x, obj1.y-obj2.y)

def obj_heading(obj, target):
    from math import atan2, pi
    return atan2(target.x-obj.x,target.y-obj.y)*180.0/pi +90

class Ant:
    def __init__(self,x,y,O, queen, color, resources= 50.0, life =  50.0, 
            state = EXPLORE, target = None):
        import random
        if color == RED:
            self.sprite = cocos.sprite.Sprite('data/ant.png')
        elif color == BLUE:
            self.sprite = cocos.sprite.Sprite('data/blueant.png')
        elif color == GREEN:
            self.sprite = cocos.sprite.Sprite('data/greenant.png')
        self.sprite.scale = 0.5
        self.x = x
        self.y = y
        self.O = O
        self.speed = 25.0
        self.resources = resources
        self.max_resources = 200.0
        self.life = life
        self.max_life = 100
        self.dps = 20.0
        self.state = EXPLORE
        self.target = target
        self.queen = queen
        self.last_think = random.random()
        ANT_HASHMAP.insert((self.x, self.y),self)
        self.old_x = x
        self.old_y = y
        self.color=color

        self.next_mutation = FlyingAnt

    def __str__(self):
        #return 'Ant: color = %s, life = %.0f, state = %s'%(self.color,self.life,self.state)
        return 'Ant - Life: %.0f Food: %.0f'%(self.life, self.resources)

    def do(self, dt):
        self.last_think += dt
        if self.last_think > 1.0:
            self.think()
            self.last_think -= 1.0

        t = m.get_terrain(int(self.x), int(self.y))
        t_speed = terrain_speed[t]
        
        if self.state == EXPLORE:
            self.explore(dt, t_speed)

        elif self.state == GET_RESOURCE:
            if obj_dist(self.target,self) < 5.0:
                dr = 20.0 * dt
                dr = min((dr,self.max_resources-self.resources))
                dr = min(dr, self.target.resources)
                self.target.resources -= dr
                self.resources += dr
                self.think()
            else:
                self.O = obj_heading(self,self.target)
                self.move(dt, t_speed)

        elif self.state == GO_HOME:
            self.target = self.queen
            if obj_dist(self.target,self) < 5.0:
                dr = 50.0 * dt
                dr = min((dr,self.resources - 50.0))
                dr = max((dr, 0.0))
                dr = dr*0.5
                self.resources -= dr
                self.target.resources += dr
                #self.think()
            else:
                self.O = obj_heading(self,self.target)
                self.move(dt, t_speed)
                # states - exploring, getting resources, attacking

        elif self.state == FIGHTING:
            if obj_dist(self.target,self) < 5.0:
                self.target.life -= self.dps * dt
            else:
                self.O = obj_heading(self,self.target)
                self.move(dt, t_speed)
       # elif self.state == FLEEING:
       #     self.O = obj_heading(self,self.queen)
       #     self.move(dt, t_speed)

        dr = min(self.resources, dt*1)
        self.resources -= dr
        self.life += dr * 1.0
        self.life = min(self.life, self.max_life)
        dl = dt*1.0 - dr
        self.life -= dl
        #if dl > 0: print 'ouch'
        if self.life <= 0:
            ABATCH.remove(self.sprite)
            ANTS.remove(self)
            body = Resource(self.x,self.y,30.0 + self.resources, body=True)
            RBATCH.add(body.sprite)
            ANT_HASHMAP.remove((self.old_x, self.old_y),self)

    def think(self):
        # if exploring, do anything interesting
        # if near queen:
        # 
        # if fighting, fight or flee
        # if fleeing, maybe go back to exploring
        # if going home, 
        ANT_HASHMAP.move((self.x,self.y), (self.old_x,self.old_y), self)
        self.old_x,self.old_y = self.x,self.y
        import random

        #print 'think'
        if self.state in (FIGHTING, FLEEING):
            if self.target.life < 0:
                self.state = EXPLORE
                self.target = None
            #if (self.life + 1000.0/obj_dist(self,self.queen) > 50):
                #self.state = FIGHTING
            #else:
            else:
                self.state = FIGHTING

        if self.state in (GET_RESOURCE, EXPLORE, GO_HOME, FIGHTING):
            nearest_enemy = None
            nearest_enemy_score = 0.0
            for cell in ANT_HASHMAP.get((self.x,self.y)):
                for i in cell:
                    if i.queen == self.queen:
                        continue # boring
                    enemy_score = obj_dist(i,self)
                    if enemy_score > 150.0:
                        continue
                    #print 'enemy'
                    #enemy_score = enemy_score * random.uniform(0.5,1.0)
                    if nearest_enemy_score < enemy_score:
                        nearest_enemy_score = enemy_score
                        nearest_enemy = i
            if nearest_enemy is not None:
                self.target = nearest_enemy
                self.state = FIGHTING

        if self.state in (GET_RESOURCE, EXPLORE):
            if self.resources >= self.max_resources:
                self.state = GO_HOME

        if self.state == GET_RESOURCE:
            if (self.target.resources <= 0):
                self.target.remove()
                self.target = None
                self.O = random.uniform(0,360)
                self.state = EXPLORE

        if self.state == GO_HOME:
            if self.resources <= 50.0:
                self.O -= 180
                self.state = EXPLORE

        if self.state == (EXPLORE):
            # sanity?
            nearest_resource = None
            nearest_resource_score = 0.0
            for cell in RESOURCE_HASHMAP.get((self.x,self.y)):
                for i in cell:
                    resource_dist = max((obj_dist(i,self),2.))
                    if resource_dist > 200.0:
                        continue
                    #resource_score = i.resources/(resource_dist*resource_dist)
                    #resource_score = resource_score*random.uniform(0.0,1.0)
                    if nearest_resource_score < resource_dist:
                        nearest_resource_score = resource_dist
                        nearest_resource = i

            if nearest_resource is not None:
                self.state = GET_RESOURCE
                self.target = i
                #print 'target',self.target

    def move(self, dt, t_speed):
        from math import pi,sin,cos
        
        if self.O > 360.0: self.O -= 360.0
        if self.O < -360.0: self.O += 360.0
        self.x += self.speed * cos((self.O-180)*pi/180.0) * dt * t_speed
        self.y += self.speed * sin((-self.O-180)*pi/180.0) * dt * t_speed

    def explore(self,dt,t_speed):
        from random import normalvariate
        from math import sqrt
        
        dO = normalvariate(0.0,sqrt(dt))*10/t_speed
        self.O += dO
        self.move(dt, t_speed)

    def draw(self):
        self.sprite.rotation = self.O
        self.sprite.x = self.x
        self.sprite.y = self.y

    def mutate(self):
        ABATCH.remove(self.sprite)
        ANTS.remove(self)
        ANT_HASHMAP.remove((self.old_x, self.old_y),self)
        new_ant = self.next_mutation(self.x, self.y, self.O,
                self.queen, self.color,
                self.life*0.8,
                self.resources,
                state = self.state,
                target = self.target)
        ANTS.add(new_ant)
        ABATCH.add(new_ant.sprite,z=0)
        


class SoldierAnt(Ant):
    def __init__(self, x, y, O, quee, color, life, resources, state, target):
        import random
        self.sprite = cocos.sprite.Sprite('data/tankant.png')
        self.sprite.scale = 0.6
        self.x = x
        self.y = y
        self.O = O
        self.speed = 15.0
        self.resources = resources
        self.max_resources = 200.0
        self.life = life
        self.max_life = 100
        self.dps = 50.0
        self.state = state
        self.target = target
        self.queen = queen
        self.last_think = random.random()
        ANT_HASHMAP.insert((self.x, self.y),self)
        self.old_x = x
        self.old_y = y
        self.color=color
        self.next_mutation = Ant
    def __str__(self):
        #return 'Ant: color = %s, life = %.0f, state = %s'%(self.color,self.life,self.state)
        return 'Soldier - Life: %.0f Food: %.0f'%(self.life, self.resources)

class FlyingAnt(Ant):
    def __init__(self, x, y, O, quee, color, life, resources,state, target):
        import random
        self.sprite = cocos.sprite.Sprite('data/flyingant.png')
        self.sprite.scale = 0.5
        self.x = x
        self.y = y
        self.O = O
        self.speed = 20.0
        self.resources = resources
        self.max_resources = 200.0
        self.life = life
        self.max_life = 100
        self.dps = 10.0
        self.state = state
        self.target = target
        self.queen = queen
        self.last_think = random.random()
        ANT_HASHMAP.insert((self.x, self.y),self)
        self.old_x = x
        self.old_y = y
        self.color=color
        self.next_mutation = SoldierAnt

    def move(self, dt, t_speed):
        Ant.move(self, dt, 1.0)
    def __str__(self):
        #return 'Ant: color = %s, life = %.0f, state = %s'%(self.color,self.life,self.state)
        return 'Flying ant - Life: %.0f Food: %.0f'%(self.life, self.resources)

#def make_ant(team,sprite,x,y,O,speed,resources,life,):


class Queen(Ant):
    def __init__(self,x,y,O, color):
        import random
        if color == RED:
            self.sprite = cocos.sprite.Sprite('data/queen.png')
        elif color == BLUE:
            self.sprite = cocos.sprite.Sprite('data/bluequeen.png')
        elif color == GREEN:
            self.sprite = cocos.sprite.Sprite('data/greenqueen.png')
        self.sprite.scale = 1
        self.x = x
        self.y = y
        self.O = O
        self.speed = 10.0
        self.resources = 50.0
        self.max_resources = 100000.0
        self.life = 1000.0
        self.max_life = 1000.0

        self.dps = 50.0
        self.state = EXPLORE
        self.target = None
        self.queen = self # yep
        self.last_think = random.random()
        ANT_HASHMAP.insert((self.x, self.y),self)
        self.old_x = x
        self.old_y = y

        self.color=color

    def do(self, dt):
        from random import normalvariate
        from math import pi,sin,cos,sqrt

        t = m.get_terrain(int(self.x), int(self.y))
        t_speed = terrain_speed[t]

        if self.target is not None:
            if obj_dist(self.target,self) < 5.0:
                self.target = None
            else:
                self.O = obj_heading(self,self.target)

        if self.target is None:
            dO = normalvariate(0.0,sqrt(dt))*1
            self.O += dO

        if self.O > 360.0: self.O -= 360.0
        if self.O < -360.0: self.O += 360.0
        self.x += self.speed * cos((self.O-180)*pi/180.0) * dt * t_speed
        self.y += self.speed * sin((-self.O-180)*pi/180.0) * dt * t_speed

        ANT_HASHMAP.move((self.x,self.y), (self.old_x,self.old_y), self)
        self.old_x,self.old_y = self.x,self.y


        dr = min(self.resources, dt)
        self.resources -= dr
        self.life += dr
        self.life = min(self.life, self.max_life)
        dl = dt - dr
        self.life -= dl
        #if dl > 0: print 'ouch'
        if self.life <= 0:
            ABATCH.remove(self.sprite)
            ANTS.remove(self)
            body = Resource(self.x,self.y,30.0 + self.resources, body=True)
            RBATCH.add(body.sprite)
            ANT_HASHMAP.remove((self.old_x, self.old_y),self)
            for ant in ANTS:
                if ant.queen == self:
                    ant.life = -100 #dead

        #self.resources += normalvariate(1.0*dt,sqrt(dt))
        if self.resources > 200.0:
            self.resources -= 100.0
            ant = Ant(self.x,self.y,self.O, self, self.color)
            ANTS.add(ant)
            ABATCH.add(ant.sprite,z=0)
    def __str__(self):
        #return 'Ant: color = %s, life = %.0f, state = %s'%(self.color,self.life,self.state)
        return 'Queen - Life: %.0f Food: %.0f'%(self.life, self.resources)

class HashMap:
    def __init__(self, cell_size):
        self.grid = {}
        self.cell_size = float(cell_size)

    def get(self,point):
        x,y = self.round_point(point)
        for i in (-1,0,1):
            for j in (-1,0,1):
                yield self.grid.get((x+i,y+j),())

    def insert(self, point, i):
        x,y = self.round_point(point)
        try:
            self.grid[(x,y)].add(i)
        except:
            self.grid[(x,y)] = set((i,))

    def remove(self, point, i):
        x,y = self.round_point(point)
        self.grid[(x,y)].remove(i)

    def move(self, point, point0, i):
        x,y = self.round_point(point)
        x0, y0 = self.round_point(point0)
        if (x,y) == (x0,y0):
            return
        self.remove(point0, i)
        self.insert(point, i)

    def round_point(self, point):
        from math import floor
        x = floor(point[0]/self.cell_size)
        y = floor(point[1]/self.cell_size)
        return (x,y)

ANT_HASHMAP = HashMap(200)
RESOURCE_HASHMAP = HashMap(200)

queen = Queen(575,181,100,RED)
red_queen = queen

ants = [Ant(570+i,177+i,100+25*i, queen, RED) for i in range(5)]+[queen]

q2 = Queen(-1800,1000,150-180,BLUE)
ants = ants + [Ant(-1800,1000,150-180, q2, BLUE) for i in range(10)]+[q2]

q3 = Queen(1400,1400,150-180,GREEN)
ants = ants + [Ant(1400,1400,150-180, q3, GREEN) for i in range(10)]+[q3]

ANTS = set(ants)

bin_name = {
        -1:'Off the map',
        0:'Water',
        1:'Ice',
        2:'Desert',
        3:'Crops',
        4:'Forest',
        5:'City'
        }

terrain_speed = {
        -1:0.01,
        0:0.2,
        1:0.5,
        2:1.0,
        3:1.2,
        4:0.7,
        5:1.2
        }
# random ant
#ant = Ant()#cocos.sprite.Sprite('data/ant.png')
#ant.scale = 5
#cocos.director.director.ant =ant
#ant.position = (320,320)



# scent boxes - 
import cPickle
terrain_bins = cPickle.load(open('data/terrain_grid.pickle','rb'))
#resources = [[0.0 for i in row] for row in terrain_bins]

RBATCH = cocos.batch.BatchNode()
ABATCH = cocos.batch.BatchNode()
m=Map()

# make a grid
grid_size = 100
nx = m.map_width//grid_size
ny = m.map_height//grid_size
grid_x = float(m.map_width)/nx
grid_y = float(m.map_height)/ny

#resource_scent_x = [[]]

#m.add(ant.sprite,z=0)
main_scene = cocos.scene.Scene(m)

#main_scene = cocos.scene.Scene (WorldScene())
#cocos.director.director.set_show_FPS(True)
cocos.director.director.run(main_scene)

